"""
This file is performing multiple data structured files into HYPER file targeting multiple tables
"""
# package to handle files/folders and related metadata/operations
import os
# Custom classes specific to this package
from project_locale.localizations_common import LocalizationsCommon
from tableau_hyper_management.ProjectNeeds import ProjectNeeds
from tableau_hyper_management.TableauHyperApiExtraLogic import TableauHyperApiExtraLogic
from tableau_hyper_management.TypeDetermination import TypeDetermination
# get current script name
SCRIPT_NAME = os.path.basename(__file__).replace('.py', '')

# main execution logic
if __name__ == '__main__':
    # instantiate Localizations Common class
    class_lc = LocalizationsCommon()
    # ensure all localization templates files are older than translated files
    class_lc.run_localization_action('maintain_sources')
    # ensure all compiled localization files are in place (as needed for localized messages later)
    class_lc.run_localization_action('compile')
    # establish localization language to use
    language_to_use = class_lc.get_region_language_to_use_from_operating_system()
    # instantiate Extractor Specific Needs class
    class_pn = ProjectNeeds({
        'script name': SCRIPT_NAME,
        'language': language_to_use,
        'title': 'Tableau Hyper Complex Converter'
    })
    # instantiate Tableau Hyper Api Extra Logic class
    class_thael = TableauHyperApiExtraLogic(language_to_use)
    # instantiate Type Determination class
    c_td = TypeDetermination(language_to_use)
    # load JSON from specified import export sequence file
    import_export_sequence = class_pn.class_fo.fn_open_file_and_get_content(
        class_pn.parameters.import_export_sequence_file)
    for crt_ie_sequence in import_export_sequence:
        table_counter = 0
        if crt_ie_sequence['sequence-type'] == 'import-files-into-hyper':
            for crt_input in crt_ie_sequence['input-content']:
                for crt_table in crt_input['tables']:
                    action_to_apply = 'append'
                    if table_counter == 0:
                        action_to_apply = 'overwrite'
                    table_counter += 1
                    # store statistics about input file
                    class_pn.class_fo.fn_store_file_statistics({
                        'checksum included':
                            class_pn.parameters.include_checksum_in_files_statistics,
                        'file list': crt_table['input-file']['name'],
                        'file meaning': 'Input',
                        'logger': class_pn.class_ln.logger,
                        'timer': class_pn.timer,
                    })
                    # build dict for entry Data Frame build method
                    df_dict = {
                        'format': crt_table['input-file']['format'],
                        'file list': [crt_table['input-file']['name']],
                        'name': 'irrelevant',
                    }
                    if 'compression' in crt_table['input-file']:
                        df_dict['field delimiter'] = crt_table['input-file']['compression']
                    if 'field-delimiter' in crt_table['input-file']:
                        df_dict['field delimiter'] = crt_table['input-file']['field-delimiter']
                    if 'worksheet' in crt_table['input-file']:
                        df_dict['worksheet list'] = [crt_table['input-file']['worksheet']]
                    # building entry Data Frame
                    working_data_frame = class_pn.class_dio.fn_load_file_into_data_frame(
                        class_pn.class_ln.logger, class_pn.timer, df_dict)
                    # releasing memory
                    df_dict = None
                    fn_dict = {
                        'action': action_to_apply,
                        'data frame': working_data_frame,
                        'hyper file': crt_ie_sequence['output-file']['name'],
                        'input data types': class_pn.config['data_types'],
                        'input parameters': class_pn.parameters,
                        'schema name': crt_input['schema-name'],
                        'table name': crt_table['table-name'],
                    }
                    # releasing memory
                    working_data_frame = None
                    # advanced detection of data type within Data Frame
                    fn_dict['data frame structure'] = c_td.fn_get_data_frame_structure(
                        class_pn.class_ln.logger, class_pn.timer, fn_dict)
                    # determine Hyper Table Columns
                    fn_dict['hyper table columns'] = class_thael.fn_build_hyper_columns(
                        class_pn.class_ln.logger, class_pn.timer, fn_dict['data frame structure'])
                    # The rows to insert into the <hyper_table> table.
                    fn_dict['data'] = class_thael.fn_rebuild_data_frame_content_for_hyper(
                        class_pn.class_ln.logger, class_pn.timer, fn_dict)
                    # adding data to table in relevant Hyper file
                    class_thael.fn_hyper_handle(class_pn.class_ln.logger, class_pn.timer, fn_dict)
                    # store statistics about input file
                    class_pn.class_fo.fn_store_file_statistics({
                        'file list': crt_ie_sequence['output-file']['name'],
                        'file meaning': 'Output',
                        'checksum included':
                            class_pn.parameters.include_checksum_in_files_statistics,
                        'logger': class_pn.class_ln.logger,
                        'timer': class_pn.timer,
                    })
                    # releasing memory
                    fn_dict = None
                    # resetting internally used variable to avoid data bleed
                    # between different entry files and destination tables
                    class_thael.columns_for_hyper_conversion = {}
    # just final message
    class_pn.class_bn.fn_final_message(
        class_pn.class_ln.logger, class_pn.parameters.output_log_file,
        class_pn.timer.timers.total(SCRIPT_NAME))



"""
Class Parameter Handling

Facilitates handling parameters values
"""
# package to handle date and times
from datetime import datetime, timedelta
# package to allow year and/or month operations based on a reference date
import datedelta
# package to add support for multi-language (i18n)
import gettext
# package to perform mathematical operations
import math
# package to handle files/folders and related metadata/operations
import os
# package regular expressions
import re


class ParameterHandling:
    known_expressions = {
        'year': ['CY', 'CurrentYear'],
        'semester': ['CYCS', 'CurrentYearCurrentSemester'],
        'semester_week': ['CYCSCW', 'CurrentYearCurrentSemesterCurrentWeek'],
        'just_semester': ['CS', 'CurrentSemester'],
        'quarter': ['CYCQ', 'CurrentYearCurrentQuarter'],
        'quarter_week': ['CYCQCW', 'CurrentYearCurrentQuarterCurrentWeek'],
        'just_quarter': ['CQ', 'CurrentQuarter'],
        'fiscal_period': ['CYCFP', 'CurrentYearCurrentFiscalPeriod'],
        'just_fiscal_period': ['CFP', 'FiscalPeriod'],
        'month': ['CYCM', 'CurrentYearCurrentMonth'],
        'just_month': ['CM', 'CurrentMonth'],
        'week': ['CYCW', 'CurrentYearCurrentWeek'],
        'just_week': ['CW', 'CurrentWeek'],
        'day': ['CYCMCD', 'CurrentYearCurrentMonthCurrentDay'],
        'just_day': ['CD', 'CurrentDay'],
        'hour': ['CYCMCDCH', 'CurrentYearCurrentMonthCurrentDayCurrentHour'],
    }
    output_standard_formats = {
        'year': '%Y',
        'fiscal_period': '%Y0%m',
        'just_fiscal_period': '0%m',
        'month': '%Y%m',
        'just_month': '%m',
        'day': '%Y%m%d',
        'just_day': '%d',
        'hour': '%Y%m%d%H%M%S%f',
    }
    locale = None

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep) \
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts) - 1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    def build_parameters(self, local_logger, query_session_parameters, in_parameter_rules
                         , in_start_iso_weekday):
        local_logger.debug(self.locale.gettext('Seen Parameters are: {parameters}')
                           .replace('{parameters}', str(query_session_parameters)))
        parameters_type = type(query_session_parameters)
        local_logger.debug(self.locale.gettext('Parameters type is {parameter_type}')
                           .replace('{parameter_type}', str(parameters_type)))
        tp = None
        if parameters_type == dict:
            tp = tuple(query_session_parameters.values())
        elif parameters_type == list:
            tp = tuple(query_session_parameters)
        else:
            local_logger.error(self.locale.gettext(
                'Unexpected parameter type, either Dictionary or List expected, '
                + 'but seen is {parameter_type}')
                               .replace('{parameter_type}', str(parameters_type)))
            exit(1)
        local_logger.debug(self.locale.gettext(
            'Initial Tuple for Parameters is: {parameters_tuple}')
                           .replace('{parameters_tuple}', str(tp)))
        return self.stringify_parameters(local_logger, tp, in_parameter_rules, in_start_iso_weekday)

    @staticmethod
    def calculate_date_deviation(in_date, deviation_type, expression_parts):
        if expression_parts[2] is None:
            expression_parts[2] = 0
        final_dates = {
            'year': in_date + datedelta.datedelta(years=int(expression_parts[2])),
            'semester': in_date + datedelta.datedelta(months=int(expression_parts[2]) * 6),
            'semester_week': in_date + timedelta(weeks=int(expression_parts[2])),
            'quarter': in_date + datedelta.datedelta(months=int(expression_parts[2]) * 3),
            'quarter_week': in_date + timedelta(weeks=int(expression_parts[2])),
            'fiscal_period': in_date + datedelta.datedelta(months=int(expression_parts[2])),
            'month': in_date + datedelta.datedelta(months=int(expression_parts[2])),
            'week': in_date + timedelta(weeks=int(expression_parts[2])),
            'day': in_date + timedelta(days=int(expression_parts[2])),
            'hour': in_date + timedelta(hours=int(expression_parts[2])),
        }
        return final_dates.get(deviation_type)

    def calculate_date_from_expression(self, local_logger, expression_parts, in_start_iso_weekday):
        final_string = ''
        all_known_expressions = self.get_flattened_known_expressions()
        if expression_parts[1] in all_known_expressions:
            str_expression = '_'.join(expression_parts)
            local_logger.debug(self.locale.gettext(
                'A known expression "{expression_parts}" has to be interpreted')
                               .replace('{expression_parts}', str_expression))
            final_string = self.interpret_known_expression(
                datetime.now(), expression_parts, in_start_iso_weekday)
            local_logger.debug(self.locale.gettext(
                'Known expression "{expression_parts}" has been interpreted as {final_string}')
                               .replace('{expression_parts}', str_expression)
                               .replace('{final_string}', final_string))
        else:
            local_logger.debug(self.locale.gettext(
                'Unknown expression encountered: provided was "{given_expression}" '
                + 'which is not among known ones: "{all_known_expressions}"')
                               .replace('{given_expression}', str(expression_parts[1]))
                               .replace('{all_known_expressions}',
                                        '", "'.join(all_known_expressions)))
            exit(1)
        return final_string

    def eval_expression(self, local_logger, crt_parameter, in_start_iso_weekday):
        value_to_return = crt_parameter
        reg_ex = re.search(r'(CalculatedDate_[A-Za-z]{2,75}_*(-*)[0-9]{0,2})', crt_parameter)
        if reg_ex:
            parameter_value_parts = reg_ex.group().split('_')
            calculated = self.calculate_date_from_expression(
                local_logger, parameter_value_parts, in_start_iso_weekday)
            value_to_return = re.sub(reg_ex.group(), calculated, crt_parameter)
            local_logger.debug(self.locale.gettext(
                'Current Parameter is STR and has been re-interpreted as value: "{str_value}"')
                               .replace('{str_value}', str(value_to_return)))
        return value_to_return

    def get_child_parent_expressions(self):
        child_parent_values = {}
        for current_expression_group in self.known_expressions.items():
            for current_expression in current_expression_group[1]:
                child_parent_values[current_expression] = current_expression_group[0]
        return child_parent_values

    def get_flattened_known_expressions(self):
        flat_values = []
        for crt_list in self.known_expressions.values():
            flat_values += crt_list
        return flat_values

    @staticmethod
    def get_week_number_as_two_digits_string(in_date, in_start_iso_weekday=1):
        if in_start_iso_weekday == 7:
            in_date = in_date + timedelta(days=1)
        week_iso_num = datetime.isocalendar(in_date)[1]
        iso_wek_to_return = str(week_iso_num)
        if week_iso_num < 10:
            iso_wek_to_return = '0' + iso_wek_to_return
        return str(datetime.isocalendar(in_date)[0]), iso_wek_to_return

    def handle_query_parameters(self, local_logger, given_session, in_start_iso_weekday):
        tp = None
        if 'parameters' in given_session:
            parameter_rules = []
            if 'parameters-handling-rules' in given_session:
                parameter_rules = given_session['parameters-handling-rules']
            tp = self.build_parameters(
                local_logger, given_session['parameters'], parameter_rules,
                in_start_iso_weekday)
        return tp

    def interpret_known_expression(self, ref_date, expression_parts, in_start_iso_weekday):
        child_parent_expressions = self.get_child_parent_expressions()
        deviation_original = child_parent_expressions.get(expression_parts[1])
        deviation = deviation_original.replace('just_', '')
        finalized_date = ref_date
        if len(expression_parts) > 2:
            finalized_date = self.calculate_date_deviation(ref_date, deviation, expression_parts)
        if deviation_original in ('just_semester', 'just_quarter', 'just_week',
                                  'semester', 'semester_week', 'quarter', 'quarter_week', 'week'):
            year_number_string, week_number_string = self.get_week_number_as_two_digits_string(
                finalized_date, in_start_iso_weekday)
            # to ensure less than 4 days within last week of year r reported to next year (ISO)
            semester_year_string = str(math.ceil(int(week_number_string) / 26))
            quarter_string = str(min(int(week_number_string),
                                     math.ceil(int(datetime.strftime(finalized_date, '%m')) / 3)))
            # values pre-calculated
            values_to_pick = {
                'just_semester': 'H' + semester_year_string,
                'just_quarter': 'Q' + quarter_string,
                'just_week': week_number_string,
                'semester': str(datetime.strftime(finalized_date, '%Y'))
                            + 'H' + semester_year_string,
                'semester_week': year_number_string + 'H' + semester_year_string
                                 + 'wk' + week_number_string,
                'quarter': str(datetime.strftime(finalized_date, '%Y')) + 'Q' + quarter_string,
                'quarter_week': year_number_string + 'Q' + quarter_string
                                + 'wk' + week_number_string,
                'week': year_number_string + week_number_string,
            }
            final_string = values_to_pick.get(deviation_original)
        else:
            final_string = datetime.strftime(
                finalized_date, self.output_standard_formats.get(deviation_original))
        return final_string

    def manage_parameter_value(self, in_logger, in_prefix, in_parameter, in_parameter_rules):
        element_to_join = ''
        if in_prefix == 'dict':
            element_to_join = in_parameter.values()
        elif in_prefix == 'list':
            element_to_join = in_parameter
        in_logger.debug(self.locale.gettext(
            'Current Parameter is {parameter_type} and has the value: "{str_value}"')
                        .replace('{parameter_type}', self.locale.gettext(in_prefix.upper()))
                        .replace('{str_value}', str(element_to_join)))
        return in_parameter_rules[in_prefix + '-values-prefix'] \
               + in_parameter_rules[in_prefix + '-values-glue'].join(element_to_join) \
               + in_parameter_rules[in_prefix + '-values-suffix']

    @staticmethod
    def set_default_starting_weekday(in_dict):
        week_starts_with_iso_weekday = 1
        if 'start-iso-weekday' in in_dict['session']:
            if in_dict['session']['start-iso-weekday'] == 'inherit-from-parent':
                in_dict['session']['start-iso-weekday'] = in_dict['query']['start-iso-weekday']
            elif in_dict['session']['start-iso-weekday'] == 'inherit-from-grand-parent':
                in_dict['session']['start-iso-weekday'] = in_dict['sequence']['start-iso-weekday']
            week_starts_with_iso_weekday = in_dict['session']['start-iso-weekday']
        return week_starts_with_iso_weekday

    def simulate_final_query(self, local_logger, timer, in_query, in_parameters_number, in_tp):
        timer.start()
        return_query = in_query
        if in_parameters_number > 0:
            try:
                return_query = in_query % in_tp
            except TypeError as e:
                local_logger.error(self.locale.gettext(
                    'Initial query expects {expected_parameters_counted} '
                    + 'but only {given_parameters_counted} were provided')
                                   .replace('{expected_parameters_counted}',
                                            str(in_parameters_number))
                                   .replace('{given_parameters_counted}', str(len(in_tp))))
                local_logger.error(e)
        timer.stop()
        return return_query

    def stringify_parameters(self, local_logger, tuple_parameters, given_parameter_rules,
                             in_start_iso_weekday):
        working_list = []
        for ndx, crt_parameter in enumerate(tuple_parameters):
            current_parameter_type = type(crt_parameter)
            working_list.append(ndx)
            if current_parameter_type == str:
                local_logger.debug(self.locale.gettext(
                    'Current Parameter is {parameter_type} and has the value: "{str_value}"')
                                   .replace('{parameter_type}', self.locale.gettext('STR'))
                                   .replace('{str_value}', crt_parameter))
                working_list[ndx] = self.eval_expression(
                    local_logger, crt_parameter, in_start_iso_weekday)
            elif current_parameter_type in (list, dict):
                prefix = str(current_parameter_type).replace("<class '", '').replace("'>", '')
                working_list[ndx] = self.manage_parameter_value(
                    local_logger, prefix.lower(), crt_parameter, given_parameter_rules)
        final_tuple = tuple(working_list)
        local_logger.debug(self.locale.gettext('Final Tuple for Parameters is: {final_tuple}')
                           .replace('{final_tuple}', str(final_tuple)))
        return final_tuple



"""
main - entry point of the package

This file is connecting to a Tableau Server and publishes a local HYPER file
measuring time elapsed (performance)
"""
# package to handle files/folders and related metadata/operations
import os

# Custom classes specific to this package
from project_locale.localizations_common import LocalizationsCommon
from tableau_hyper_management.ProjectNeeds import ProjectNeeds
from tableau_hyper_management.TableauServerCommunicator import TableauServerCommunicator
# get current script name
SCRIPT_NAME = os.path.basename(__file__).replace('.py', '')

# main execution logic
if __name__ == '__main__':
    # instantiate Localizations Common class
    class_lc = LocalizationsCommon()
    # ensure all localization templates files are older than translated files
    class_lc.run_localization_action('maintain_sources')
    # ensure all compiled localization files are in place (as needed for localized messages later)
    class_lc.run_localization_action('compile')
    # establish localization language to use
    language_to_use = class_lc.get_region_language_to_use_from_operating_system()
    # instantiate Extractor Specific Needs class
    class_pn = ProjectNeeds({
        'script name': SCRIPT_NAME,
        'language': language_to_use,
        'title': 'Tableau Data Source Publisher'
    })
    # as input and/or output file might contain CalculatedDate expression
    # an evaluation is required
    class_pn.parameters.input_file = class_pn.class_ph.eval_expression(
        class_pn.class_ln.logger, class_pn.parameters.input_file, 7)
    # validate input file existence
    class_pn.class_bn.fn_validate_single_value(class_pn.parameters.input_file, 'file')
    # get the input file into a list
    relevant_files_list = class_pn.class_fo.fn_build_file_list(
            class_pn.class_ln.logger, class_pn.timer, class_pn.parameters.input_file)
    # log file statistic details
    class_pn.class_fo.fn_store_file_statistics({
        'file list': relevant_files_list,
        'file meaning': 'Input',
        'checksum included': class_pn.parameters.include_checksum_in_files_statistics,
        'logger': class_pn.class_ln.logger,
        'timer': class_pn.timer,
    })
    # get the secrets from provided file
    credentials = class_pn.class_fo.fn_open_file_and_get_content(
            class_pn.parameters.input_credentials_file, 'json')
    credentials_dict = credentials['Credentials']['LDAP']['Production']['Default']
    # instantiate main library that ensure Tableau Server communication
    c_tsc = TableauServerCommunicator(language_to_use)
    # initiate Tableau Server connection
    c_tsc.connect_to_tableau_server(class_pn.class_ln.logger, class_pn.timer, {
        'Tableau Server': class_pn.parameters.tableau_server,
        'Tableau Site': class_pn.parameters.tableau_site,
        'Username': credentials_dict['Username'],
        'Password': credentials_dict['Password'],
    })
    # identify relevant project(s) based on given name
    list_project_details = c_tsc.load_tableau_project_ids(
            class_pn.class_ln.logger, class_pn.timer,
            [class_pn.parameters.tableau_project], 'JustOnesMentioned')
    # check if a single project has been identified and if so proceed with publishing
    if c_tsc.is_publishing_possible(
            class_pn.class_ln.logger, class_pn.parameters.tableau_project, list_project_details):
        for current_input_file in relevant_files_list:
            # perform the publishing of data source
            c_tsc.publish_data_source_to_tableau_server(class_pn.class_ln.logger, class_pn.timer, {
                'Project ID': list_project_details[0],
                'Tableau Extract File': current_input_file,
                'Publishing Mode': class_pn.parameters.publishing_mode,
            })
    # disconnect from Tableau Server
    c_tsc.disconnect_from_tableau_server(class_pn.class_ln.logger, class_pn.timer)
    # just final message
    class_pn.class_bn.fn_final_message(
            class_pn.class_ln.logger, class_pn.parameters.output_log_file,
            class_pn.timer.timers.total(SCRIPT_NAME))



"""
Class Converter Specific Needs

Handling specific needs for Extractor script
"""
# useful methods to measure time performance by small pieces of code
from codetiming import Timer
# package to add support for multi-language (i18n)
import gettext
# package to facilitate operating system operations
import os
# package to facilitate common operations
from .BasicNeeds import BasicNeeds
from .CommandLineArgumentsManagement import CommandLineArgumentsManagement
from .DataInputOutput import DataInputOutput
from .FileOperations import FileOperations
from .ParameterHandling import ParameterHandling
from .LoggingNeeds import LoggingNeeds


class ProjectNeeds:
    class_bn = None
    class_clam = None
    class_dio = None
    class_fo = None
    class_ph = None
    config = None
    locale = None
    parameters = None
    script = None
    timer = None

    def __init__(self, in_dict):
        self.script = in_dict['script name']
        self.initiate_locale(in_dict['language'])
        self.initiate_classes(in_dict['language'])
        # load application configuration (inputs are defined into a json file)
        self.load_configuration()
        # adding a special case data type
        self.config['data_types']['empty'] = '^$'
        self.config['data_types']['str'] = ''
        # initiate Logging sequence
        self.initiate_logger_and_timer()
        # reflect title and input parameters given values in the log
        self.class_clam.listing_parameter_values(
            self.class_ln.logger, self.timer, in_dict['title'],
            self.config['input_options'][self.script], self.parameters)

    def fn_check_inputs_specific(self, input_parameters):
        if self.script == 'publisher':
            self.class_bn.fn_timestamped_print(self.locale.gettext(
                'Checking if provided input credentials file exists'))
            self.class_bn.fn_validate_single_value(input_parameters.input_credentials_file, 'file')
            self.class_bn.fn_timestamped_print(self.locale.gettext(
                'Checking if provided Tableau Server url is valid'))
            self.class_bn.fn_validate_single_value(input_parameters.tableau_server, 'url')

    def initiate_classes(self, default_language):
        # instantiate Basic Needs class
        self.class_bn = BasicNeeds(default_language)
        # instantiate File Operations class
        self.class_fo = FileOperations(default_language)
        # instantiate Command Line Arguments class
        self.class_clam = CommandLineArgumentsManagement(default_language)
        # instantiate Data Manipulator class, useful to manipulate data frames
        self.class_dio = DataInputOutput(default_language)
        # instantiate Logger class
        self.class_ln = LoggingNeeds()
        # instantiate Parameter Handling class
        self.class_ph = ParameterHandling(default_language)

    def initiate_locale(self, default_language):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(
            locale_domain, localedir=locale_folder, languages=[default_language], fallback=True)

    def initiate_logger_and_timer(self):
        # initiate logger
        self.class_ln.initiate_logger(self.parameters.output_log_file, self.script)
        # initiate localization specific for this script
        # define global timer to use
        self.timer = Timer(self.script,
                           text=self.locale.gettext('Time spent is {seconds}'),
                           logger=self.class_ln.logger.debug)

    def load_configuration(self):
        # load application configuration (inputs are defined into a json file)
        ref_folder = os.path.dirname(__file__).replace('tableau_hyper_management', 'config')
        config_file = os.path.join(ref_folder, 'tableau-hyper-management.json').replace('\\', '/')
        self.config = self.class_fo.fn_open_file_and_get_content(config_file)
        # get command line parameter values
        self.parameters = self.class_clam.parse_arguments(self.config['input_options'][self.script])
        # checking inputs, if anything is invalid an exit(1) will take place
        self.class_bn.fn_check_inputs(self.parameters)
        # checking inputs, if anything is invalid an exit(1) will take place
        self.fn_check_inputs_specific(self.parameters)

    def listing_parameter_values(self, in_logger, timer, title, in_config, given_parameter_values):
        timer.start()
        in_logger.info('=' * 50)
        in_logger.info(self.locale.gettext('{application_name} has started')
                       .replace('{application_name}', title))
        in_logger.info('~' * 50)
        in_logger.info(self.locale.gettext('Overview of input parameter given values'))
        in_logger.info('~' * 50)
        parameter_values_dictionary = given_parameter_values.__dict__
        for input_key, attributes in in_config.items():
            # checking first if short key was provided, otherwise consider longer
            if input_key in parameter_values_dictionary:
                key_value_to_consider = input_key
            else:
                key_value_to_consider = attributes['option_long'].replace('-', '_')
            # having the key consider we determine the value of the current parameter
            value_to_consider = parameter_values_dictionary[key_value_to_consider]
            # we build the parameter feedback considering "option_description"
            # and replacing %s with parameter value
            feedback = self.locale.gettext(attributes['option_description']) \
                .replace('%s', value_to_consider)
            # we finally write the feedback to logger
            in_logger.info(feedback)
        in_logger.info('~' * 50)
        timer.stop()

    def source_vs_destination_file_modification_assesment(self, in_logger, timer, in_dict):
        """
        Checks if any source file is newer than destination

        :param in_logger: logger handler to capture running details
        :param timer: pointer to measure code performance
        :param in_dict: dictionary containing following keys with relevant values:
            "destination file" and "list source files"
        :return: either None or localized term for "newer"
        """
        timer.start()
        destination_file_times = self.class_fo.fn_get_file_dates_raw(in_dict['destination file'])
        source_files_count = len(in_dict['list source files'])
        loop_counter = 0
        stop_verdict = self.class_fo.locale.gettext('newer')
        verdict = None
        while verdict != stop_verdict and loop_counter < source_files_count:
            crt_file = in_dict['list source files'][loop_counter]
            crt_verdict = self.class_fo.fn_get_file_datetime_verdict(
                in_logger, crt_file, 'last modified', destination_file_times['last modified'])
            if crt_verdict == stop_verdict:
                verdict = crt_verdict
            loop_counter += 1
        timer.stop()
        return verdict



"""
DataInput - class to handle data storing to disk (from Pandas Data Frame
"""


class DataDiskWrite:
    implemented_disk_write_file_types = ['csv', 'excel', 'json', 'parquet', 'pickle']

    @staticmethod
    def fn_internal_store_data_frame_to_csv_file(in_dict):
        if in_dict['format'].lower() == 'csv':
            try:
                in_dict['in data frame'].to_csv(
                    path_or_buf=in_dict['name'], sep=in_dict['field delimiter'],
                    header=True, index=False, encoding='utf-8')
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_store_data_frame_to_excel_file(in_dict):
        if in_dict['format'].lower() == 'excel':
            try:
                in_dict['in data frame'].to_excel(
                    excel_writer=in_dict['name'], engine='xlsxwriter',
                    freeze_panes=(1, 1), encoding='utf-8', index=False, verbose=False)
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_store_data_frame_to_json_file(in_dict):
        if in_dict['format'].lower() == 'json':
            try:
                in_dict['in data frame'].to_json(
                    path_or_buf=in_dict['name'], compression=in_dict['compression'])
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_store_data_frame_to_parquet_file(in_dict):
        if in_dict['format'].lower() == 'parquet':
            try:
                in_dict['in data frame'].to_parquet(
                    path=in_dict['name'], compression=in_dict['compression'],
                    use_deprecated_int96_timestamps=True)
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_store_data_frame_to_pickle_file(in_dict):
        if in_dict['format'].lower() == 'pickle':
            try:
                in_dict['in data frame'].to_pickle(
                    path=in_dict['name'], compression=in_dict['compression'])
            except Exception as err:
                in_dict['error details'] = err
        return in_dict



"""
localizations_common - common class for various localizations tasks
"""
# package to handle date and times
from datetime import datetime
# package to add support for multi-language (i18n)
import gettext
# glob package
import glob
# package to facilitate operating system project_locale detection
import locale
# package to handle files/folders and related metadata/operations
import os
import pathlib
# package to facilitate multiple operation system operations
import platform


class LocalizationsCommon:
    locale = None
    locale_implemented = [
        'de_DE',
        'es_ES',
        'fr_FR',
        'it_IT',
        'ro_RO',
    ]
    operation_is_required = False

    def __init__(self, in_language='en_US'):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep)\
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts)-1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts), 'project_locale'), locale_domain))
        self.locale = gettext.translation(
            locale_domain, localedir=locale_folder, languages=[in_language], fallback=True)

    def check_file_pairs(self, in_dict):
        get_details_to_operate = False
        file_situation_verdict = ''
        if os.path.lexists(pathlib.Path(in_dict['destination'])):
            source_last_modified = os.path.getmtime(in_dict['source'])
            destination_last_modified = os.path.getmtime(in_dict['destination'])
            if source_last_modified > destination_last_modified:
                self.operation_is_required = True
                get_details_to_operate = True
                print(self.locale.gettext(
                    '#{file_counter}, For locale {locale_name} the {source_file_type} file '
                    + '{source_file_name} is newer than compiled one {destination_file_name}, '
                    + 'therefore to remedy this, {destination_operation_name} is required')
                      .replace('{file_counter}', str(in_dict['counter']))
                      .replace('{locale_name}', in_dict['locale'])
                      .replace('{source_file_type}', in_dict['source file type name'])
                      .replace('{source_file_name}', os.path.basename(in_dict['source']))
                      .replace('{destination_file_name}', os.path.basename(in_dict['destination']))
                      .replace('{destination_operation_name}',
                               in_dict['destination operation name']))
                print(self.locale.gettext('===> source file {source_file_name} '
                                          + 'has {source_last_modified} '
                                          + 'vs. {destination_last_modified} of the '
                                          + 'destination file {destination_file_name}')
                      .replace('{source_file_name}', os.path.basename(in_dict['source']))
                      .replace('{source_last_modified}',
                           str(datetime.fromtimestamp(source_last_modified)))
                      .replace('{destination_last_modified}',
                           str(datetime.fromtimestamp(destination_last_modified)))
                      .replace('{destination_file_name}', os.path.basename(in_dict['destination'])))
                file_situation_verdict = 'newer'
        else:
            self.operation_is_required = True
            get_details_to_operate = True
            print(self.locale.gettext(
                '#{file_counter}, For locale {locale_name} the file {source_file_name}'
                + 'is missing, therefore to remedy this, '
                + '{destination_operation_name} is required into following folder: '
                + '{destination_folder}')
                  .replace('{file_counter}', str(in_dict['counter']))
                  .replace('{locale_name}', in_dict['locale'])
                  .replace('{source_file_name}', os.path.basename(in_dict['source']))
                  .replace('{destination_operation_name}', in_dict['destination operation name'])
                  .replace('{destination_folder}',
                           os.path.normpath(os.path.dirname(in_dict['destination']))))
            file_situation_verdict = 'missing'
        return get_details_to_operate, file_situation_verdict

    @staticmethod
    def file_counter_limit(in_file_counter, in_file_list_size):
        file_list_paring_complete = False
        if in_file_counter == in_file_list_size:
            file_list_paring_complete = True
        return file_list_paring_complete

    def get_region_language_to_use_from_operating_system(self):
        language_to_use = 'en_US'
        try:
            region_language_to_use = locale.getdefaultlocale('LC_ALL')
            if region_language_to_use[0] in self.locale_implemented:
                language_to_use = region_language_to_use[0]
        except AttributeError as err:
            print(err)
        except ValueError as err:
            print(err)
        return language_to_use

    @staticmethod
    def get_this_file_folder():
        return os.path.dirname(__file__)

    def get_project_localisation_source_files(self, in_extension):
        file_pattern = os.path.join(os.path.dirname(__file__), '**/*.' + in_extension)
        initial_list = glob.glob(file_pattern, recursive=True)
        normalizer = lambda x: self.path_normalize(x)
        return list(map(normalizer, initial_list))

    @staticmethod
    def get_project_root():
        file_parts = os.path.normpath(os.path.abspath(__file__))\
            .replace('\\', os.altsep).split(os.altsep)
        return os.altsep.join(file_parts[:-3])

    def get_virtual_environment_python_binary(self):
        python_binary = 'python'
        if platform.system() == 'Windows':
            project_root = self.get_project_root()
            join_separator = os.path.altsep
            elements_to_join = [
                project_root,
                'virtual_environment',
                'Scripts',
            ]
            virtual_env_long = join_separator.join(elements_to_join)
            elements_to_join[1] = 'venv'
            virtual_env_short = join_separator.join(elements_to_join)
            if os.path.isdir(virtual_env_long):
                python_binary = os.path.join(virtual_env_long, 'python.exe')
            elif os.path.isdir(virtual_env_short):
                python_binary = os.path.join(virtual_env_short, 'python.exe')
        return os.path.normpath(python_binary)

    def operate_localisation_files(self, in_dict_details_to_operate_with):
        if self.operation_is_required:
            for current_details_to_operate in in_dict_details_to_operate_with:
                os.system(self.get_virtual_environment_python_binary() + ' '
                          + os.path.join(os.path.dirname(__file__), 'localizations_setup.py')
                          + ' ' + current_details_to_operate['operation']
                          + ' --input-file=' + current_details_to_operate['input-file']
                          + ' --output-file=' + current_details_to_operate['output-file']
                          + ' --locale ' + current_details_to_operate['locale']
                          + current_details_to_operate['operation final flags'])

    @staticmethod
    def path_normalize(in_file_name):
        return os.path.join(os.path.normpath(os.path.dirname(in_file_name)),
                            os.path.basename(in_file_name))

    def run_localization_action(self, in_action):
        known_actions = {
            'compile': 'localizations_compile.py',
            'maintain_sources': 'localizations_maintain_sources.py',
        }
        command_parts_to_run = [
            self.get_virtual_environment_python_binary(),
            ' ',
            self.get_this_file_folder(),
            os.path.altsep,
            known_actions.get(in_action),
        ]
        command_line = ''.join(command_parts_to_run)
        print('Executing "' + command_line + '"...')
        os.system(command_line)



from datetime import datetime
import os
from db_extractor.LoggingNeeds import LoggingNeeds
from db_extractor.FileOperations import FileOperations
import unittest
# package to facilitate multiple operation system operations
import platform


class TestFileOperations(unittest.TestCase):

    def setUp(self) -> None:
        python_binary = 'python'
        if platform.system() == 'Windows':
            python_binary += '.exe'
        os.system(python_binary + ' '
                  + os.path.join(os.path.normpath(os.path.dirname(__file__))
                                 .replace('test', 'sources/project_locale'),
                                 'localizations_compile.py'))

    def test_file_statistics(self):
        class_fo = FileOperations()
        value_to_assert = class_fo.fn_get_file_statistics({'file name': __file__})['size [bytes]']
        value_to_compare_with = os.path.getsize(__file__)
        self.assertEqual(value_to_assert, value_to_compare_with)

    def test_file_dates(self):
        class_fo = FileOperations()
        value_to_assert = class_fo.fn_get_file_dates(__file__)['created']
        value_to_compare_with = datetime.fromtimestamp(os.path.getctime(__file__))
        self.assertEqual(value_to_assert, value_to_compare_with)

    def test_file_verdict(self):
        class_ln = LoggingNeeds()
        class_ln.initiate_logger('None', __file__)
        class_fo = FileOperations()
        ref_datetime = os.path.getctime(__file__)
        relevant_file = os.path.join(
            os.path.dirname(__file__).replace('test', 'sources'), 'project_locale/__init__.py')
        value_to_assert = class_fo.fn_get_file_datetime_verdict(
            class_ln.logger, relevant_file, 'created', ref_datetime)
        self.assertEqual(value_to_assert, class_fo.locale.gettext('older'))



"""
TableauHyperApiExtraLogic - a Hyper client library.

This library allows packaging CSV content into HYPER format with data type checks
"""
# package to add support for multi-language (i18n)
import gettext
# package to handle files/folders and related metadata/operations
import os
# package regular expression
import re
# package to handle numerical structures
import numpy
# package to handle Data Frames (in this file)
import pandas as pd
# Custom classes from Tableau Hyper package
from tableauhyperapi import HyperProcess, Telemetry, Connection, CreateMode, \
    NOT_NULLABLE, NULLABLE, SqlType, TableDefinition, TableName, Inserter, HyperException


class TableauHyperApiExtraLogic:
    locale = None
    supported_input_file_types = ('csv', 'json', 'parquet', 'pickle')
    supported_output_file_types = ('csv', 'parquet', 'pickle')
    columns_for_hyper_conversion = {}
    hyper_conversion_dtypes = ['str', 'int64', 'float']

    def __init__(self, in_language):
        file_parts = os.path.normpath(os.path.abspath(__file__)).replace('\\', os.path.altsep) \
            .split(os.path.altsep)
        locale_domain = file_parts[(len(file_parts) - 1)].replace('.py', '')
        locale_folder = os.path.normpath(os.path.join(
            os.path.join(os.path.altsep.join(file_parts[:-2]), 'project_locale'), locale_domain))
        self.locale = gettext.translation(locale_domain, localedir=locale_folder,
                                          languages=[in_language], fallback=True)

    def fn_build_hyper_columns(self, logger, timer, in_data_frame_structure):
        timer.start()
        list_to_return = []
        for current_field_structure in in_data_frame_structure:
            list_to_return.append(current_field_structure['order'])
            current_column_type = self.fn_convert_to_hyper_types(current_field_structure['type'])
            logger.debug(self.locale.gettext(
                'Column {column_order} having name [{column_name}] and type "{column_type}" '
                + 'will become "{column_type_new}"')
                         .replace('{column_order}', str(current_field_structure['order']))
                         .replace('{column_name}', current_field_structure['name'])
                         .replace('{column_type}', str(current_field_structure['type']))
                         .replace('{column_type_new}', str(current_column_type)))
            nullability_value = NULLABLE
            if current_field_structure['nulls'] == 0:
                nullability_value = NOT_NULLABLE
            list_to_return[current_field_structure['order']] = TableDefinition.Column(
                name=current_field_structure['name'],
                type=current_column_type,
                nullability=nullability_value
            )
        logger.info(self.locale.gettext('Building Hyper columns completed'))
        timer.stop()
        return list_to_return

    def fn_convert_multiple_columns(self, in_logger, timer, in_data_frame, in_target_dtype):
        # if there's a list of columns to be converted to Integer do that
        if in_target_dtype in self.columns_for_hyper_conversion:
            timer.start()
            col_to_convert = self.columns_for_hyper_conversion[in_target_dtype]
            if in_target_dtype in ('int64', 'float'):
                in_data_frame[col_to_convert] = in_data_frame[col_to_convert]\
                    .fillna(0).astype(in_target_dtype)
            else:
                in_data_frame[col_to_convert] = in_data_frame[col_to_convert]\
                    .astype(in_target_dtype)
            in_logger.info(self.locale.gettext(
                'Converting multiple columns "{column_list}" to {target_data_type} completed')
                           .replace('{column_list}', '", "'.join(col_to_convert))
                           .replace('{target_data_type}', in_target_dtype.upper()))
            timer.stop()
        return in_data_frame

    @staticmethod
    def fn_convert_to_hyper_types(given_type):
        switcher = {
            'empty': SqlType.text(),
            'bool': SqlType.bool(),
            'int': SqlType.big_int(),
            'float-dot': SqlType.double(),
            'date-YMD': SqlType.date(),
            'date-MDY': SqlType.date(),
            'date-DMY': SqlType.date(),
            'time-24': SqlType.time(),
            'time-12': SqlType.time(),
            'datetime-24-YMD': SqlType.timestamp(),
            'datetime-12-MDY': SqlType.timestamp(),
            'datetime-24-DMY': SqlType.timestamp(),
            'str': SqlType.text()
        }
        identified_type = switcher.get(given_type)
        if identified_type is None:
            identified_type = SqlType.text()
        return identified_type

    def fn_delete_data_from_hyper(self, in_logger, timer, in_dict):
        timer.start()
        in_logger.debug(self.locale.gettext(
            'Hyper SQL about to be executed is: {hyper_sql}')
                        .replace('{hyper_sql}', in_dict['sql query']))
        row_count = in_dict['connection'].execute_command(command=in_dict['query'])
        in_logger.debug(self.locale.gettext(
            'Hyper SQL executed with success and {rows_counted} have been retrieved')
                        .replace('{rows_counted}', str(row_count)))
        timer.stop()

    def fn_insert_data_into_hyper_table(self, local_logger, timer, in_dict):
        timer.start()
        # Execute the actual insert
        with Inserter(in_dict['connection'], in_dict['table']) as hyper_insert:
            hyper_insert.add_rows(rows=in_dict['data'])
            hyper_insert.execute()
        local_logger.info(self.locale.gettext('Data has been inserted into Hyper table'))
        timer.stop()

    def fn_create_hyper_schema(self, local_logger, timer, in_dict):
        timer.start()
        in_dict['connection'].catalog.create_schema(in_dict['schema name'])
        local_logger.info(self.locale.gettext(
            'Hyper schema "{hyper_schema_name}" has been created')
                          .replace('{hyper_schema_name}', in_dict['schema name']))
        timer.stop()

    def fn_create_hyper_table(self, local_logger, timer, in_dict):
        timer.start()
        out_hyper_table = TableDefinition(
            TableName(in_dict['schema name'], in_dict['table name']),
            columns=in_dict['columns'],
        )
        in_dict['connection'].catalog.create_table(table_definition=out_hyper_table)
        local_logger.info(self.locale.gettext(
            'Hyper table "{hyper_table_name}" has been created')
                          .replace('{hyper_table_name}', in_dict['table name']))
        timer.stop()
        return out_hyper_table

    def fn_get_column_names_from_table(self, in_logger, in_dict):
        columns_counted = in_dict['table definition'].column_count
        in_logger.debug(self.locale.gettext('A number of {column_count} columns were found')
                        .replace('{column_count}', str(columns_counted)))
        columns_counter = 0
        table_columns = []
        while columns_counter < columns_counted:
            table_columns.append(columns_counter)
            table_columns[columns_counter] = str(in_dict['table definition'].get_column(
                columns_counter).name).replace('"', '')
            columns_counter += 1
        in_logger.debug(self.locale.gettext('And these columns are: {column_list}')
                        .replace('{column_list}', str(table_columns)))
        return table_columns

    def fn_get_records_count_from_table(self, local_logger, timer, in_dict):
        timer.start()
        # Number of rows in the <hyper_table> table.
        # `execute_scalar_query` is for executing a query
        # that returns exactly one row with one column.
        query_to_run = 'SELECT COUNT(*) FROM "{hyper_schema_name}"."{hyper_table_name}"' \
            .replace('{hyper_schema_name}', in_dict['schema name']) \
            .replace('{hyper_table_name}', in_dict['table name'])
        row_count = in_dict['connection'].execute_scalar_query(query=query_to_run)
        local_logger.info(self.locale.gettext('Table {hyper_table_name} has {row_count} rows')
                          .replace('{hyper_schema_name}', in_dict['schema name']) \
                          .replace('{hyper_table_name}', in_dict['table name']) \
                          .replace('{row_count}', str(row_count)))
        timer.stop()

    def fn_hyper_handle(self, in_logger, timer, in_dict):
        timer.start()
        out_data_frame = None
        try:
            # Starts Hyper Process with telemetry enabled/disabled to send data to Tableau or not
            # To opt in, simply set telemetry=Telemetry.SEND_USAGE_DATA_TO_TABLEAU.
            # To opt out, simply set telemetry=Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU.
            telemetry_chosen = Telemetry.DO_NOT_SEND_USAGE_DATA_TO_TABLEAU
            with HyperProcess(telemetry=telemetry_chosen) as hyper_process:
                in_logger.debug(self.locale.gettext('Hyper engine process initialized'))
                in_logger.debug(self.locale.gettext('Chosen Telemetry is {telemetry_value}')
                                .replace('{telemetry_value}', str(telemetry_chosen)))
                timer.stop()
                timer.start()
                hyper_create_mode = {
                    'append': CreateMode.NONE,
                    'overwrite': CreateMode.CREATE_AND_REPLACE,
                    'read': CreateMode.NONE,
                    'delete': CreateMode.NONE,
                    'update': CreateMode.NONE,
                }
                #  Connect to an existing .hyper file
                with Connection(endpoint=hyper_process.endpoint,
                                database=in_dict['hyper file'],
                                create_mode=hyper_create_mode.get(in_dict['action'])
                                ) as hyper_connection:
                    in_logger.debug(self.locale.gettext(
                        'Connection to the Hyper engine using file name "{file_name}" '
                        + 'has been established')
                                    .replace('{file_name}', in_dict['hyper file']))
                    timer.stop()
                    in_dict['connection'] = hyper_connection
                    if in_dict['action'] == 'read':
                        out_data_frame = self.fn_hyper_read(in_logger, timer, in_dict)
                    elif in_dict['action'] in ('append', 'overwrite'):
                        self.fn_write_data_into_hyper_file(in_logger, timer, in_dict)
                    elif in_dict['action'] in ('delete', 'update'):
                        self.fn_delete_data_from_hyper(in_logger, timer, in_dict)
                        self.fn_get_records_count_from_table(in_logger, timer, in_dict)
            timer.start()
            hyper_connection.close()
            in_logger.info(self.locale.gettext(
                'Connection to the Hyper engine file has been closed'))
            in_logger.info(self.locale.gettext('Hyper engine process has been shut down'))
            timer.stop()
        except HyperException as ex:
            in_logger.error(str(ex).replace(chr(10), ' '))
            timer.stop()
            exit(1)
        return out_data_frame

    def fn_hyper_read(self, in_logger, timer, in_dict):
        timer.start()
        # once Hyper is opened we can get data out
        query_to_run = f"SELECT * FROM {TableName('Extract', 'Extract')}"
        in_logger.debug(self.locale.gettext(
            'Hyper SQL about to be executed is: {hyper_sql}')
                        .replace('{hyper_sql}', str(query_to_run)))
        result_set = in_dict['connection'].execute_list_query(query=query_to_run)
        out_data_frame = pd.DataFrame(result_set)
        in_logger.debug(self.locale.gettext(
            'Hyper SQL executed with success and {rows_counted} have been retrieved')
                        .replace('{rows_counted}', str(len(out_data_frame))))
        table_definition = in_dict['connection'].catalog.get_table_definition(
            name=TableName('Extract', 'Extract'))
        table_columns = self.fn_get_column_names_from_table(in_logger, {
            'table definition': table_definition,
        })
        out_data_frame.set_axis(table_columns, axis='columns', inplace=True)
        timer.stop()
        return out_data_frame

    def fn_rebuild_data_frame_content_for_hyper(self, in_logger, timer, in_dict):
        timer.start()
        in_dict['data frame'].replace(to_replace=[numpy.nan], value=[None], inplace=True)
        in_logger.info(self.locale.gettext('Filling empty values with NAN finished'))
        timer.stop()
        timer.start()
        # Cycle through all found columns
        for current_field in in_dict['data frame structure']:
            in_logger.debug(self.locale.gettext(
                'Column {column_name} has pandas data type = {column_pandas_type} '
                + 'and python data type = {column_python_type}')
                            .replace('{column_name}', current_field['name'])
                            .replace('{column_pandas_type}', str(current_field['panda_type']))
                            .replace('{column_python_type}', str(current_field['type'])))
            in_dict['data frame'][current_field['name']] = self.fn_reevaluate_single_column(
                in_dict['data frame'][current_field['name']], current_field)
        timer.stop()
        for converting_data_type in self.hyper_conversion_dtypes:
            in_dict['data frame'] = self.fn_convert_multiple_columns(
                in_logger, timer, in_dict['data frame'], converting_data_type)
        timer.start()
        in_logger.info(self.locale.gettext(
            'Re-building CSV content for maximum Hyper compatibility has been completed'))
        in_logger.info(self.locale.gettext(
            '{rows_counted} records were prepared in this process')
                       .replace('{rows_counted}', str(len(in_dict['data frame']))))
        timer.stop()
        return in_dict['data frame'].values

    def fn_reevaluate_single_column(self, df_column, in_field_details):
        if in_field_details['type'][0:5] in ('date-', 'datet', 'time-'):
            df_column = self.fn_string_to_date(df_column, in_field_details['type'])
        else:
            df_column = self.fn_reevaluate_single_column_additional(df_column, in_field_details)
        return df_column

    def fn_reevaluate_single_column_additional(self, in_df_column, in_field):
        target_data_type = self.fn_standardize_data_type(in_field)
        if target_data_type != '':
            if target_data_type in self.columns_for_hyper_conversion:
                new_index = len(self.columns_for_hyper_conversion[target_data_type])
                self.columns_for_hyper_conversion[target_data_type].append(new_index)
                self.columns_for_hyper_conversion[target_data_type][new_index] = in_field['name']
            else:
                self.columns_for_hyper_conversion[target_data_type] = [in_field['name']]
        return in_df_column

    @staticmethod
    def fn_standardize_data_type(in_field):
        target_data_type = ''
        if in_field['panda_type'] in ('object', 'float64') and in_field['type'] == 'int':
            target_data_type = 'int64'
        elif str(in_field['type']) in ('float-dot', 'str'):
            known_types = {
                'float-dot': 'float',
                'str': 'str',
            }
            target_data_type = known_types.get(in_field['type'])
        return target_data_type

    @staticmethod
    def fn_string_to_date(in_df_column, in_data_type):
        if re.match('-YMD', in_data_type):
            in_df_column = pd.to_datetime(in_df_column, yearfirst=True)
        elif re.match('-DMY', in_data_type):
            in_df_column = pd.to_datetime(in_df_column, dayfirst=True)
        else:
            in_df_column = pd.to_datetime(in_df_column)
        return in_df_column

    def fn_write_data_into_hyper_file(self, in_logger, timer, in_dict):
        if in_dict['action'] == 'append':
            try:
                self.fn_get_records_count_from_table(in_logger, timer, {
                    'connection': in_dict['connection'],
                    'schema name': in_dict['schema name'],
                    'table name': in_dict['table name'],
                })
                hyper_table = in_dict['connection'].catalog.get_table_definition(
                    TableName(in_dict['schema name'], in_dict['table name']))
            except HyperException as err:
                timer.stop()
                hyper_table = self.fn_create_hyper_table(in_logger, timer, {
                    'columns': in_dict['hyper table columns'],
                    'connection': in_dict['connection'],
                    'schema name': in_dict['schema name'],
                    'table name': in_dict['table name'],
                })
        elif in_dict['action'] == 'overwrite':
            self.fn_create_hyper_schema(in_logger, timer, in_dict)
            hyper_table = self.fn_create_hyper_table(in_logger, timer, {
                'columns': in_dict['hyper table columns'],
                'connection': in_dict['connection'],
                'schema name': in_dict['schema name'],
                'table name': in_dict['table name'],
            })
        self.fn_insert_data_into_hyper_table(in_logger, timer, {
            'connection': in_dict['connection'],
            'data': in_dict['data'],
            'table': hyper_table,
        })
        self.fn_get_records_count_from_table(in_logger, timer, {
            'connection': in_dict['connection'],
            'schema name': in_dict['schema name'],
            'table name': in_dict['table name'],
        })



"""
Testing key methods from Parameter Handling class
"""
from datetime import datetime
import os
from sources.tableau_hyper_management.FileOperations import FileOperations
from sources.tableau_hyper_management.ParameterHandling import ParameterHandling
import unittest


class TestParameterHandling(unittest.TestCase):

    def test_interpret_known_expression(self):
        class_fo = FileOperations()
        # load testing values from JSON file
        # where all cases are grouped
        json_structure = class_fo.fn_open_file_and_get_content(
            os.path.join(os.path.dirname(__file__),  'expressions.json'))
        # flatten out all testing values
        pair_values = []
        for crt_list in json_structure.values():
            pair_values += crt_list
        # parse through all pair of values and run the test
        class_ph = ParameterHandling()
        for current_pair in pair_values:
            reference_format = '%Y-%m-%d'
            if 'reference_format' in current_pair:
                reference_format = current_pair['reference_format']
            reference_date = datetime.strptime(current_pair['reference_value'], reference_format)
            expression_parts = current_pair['expression'].split('_')
            current_pair['start-iso-weekday'] = class_ph.set_default_starting_weekday({
                'session': current_pair,
                'query': current_pair,
                'sequence': current_pair,
            })
            value_to_assert = class_ph.interpret_known_expression(
                reference_date, expression_parts, current_pair['start-iso-weekday'])
            self.assertEqual(value_to_assert, current_pair['expected_value'],
                             'Provided value was "' + current_pair['reference_value']
                             + '", Expression was "' + current_pair['expression'] + '" '
                             + '", Expected was "' + current_pair['expected_value'] + '" '
                             + 'but received was "' + value_to_assert + '"...')


if __name__ == '__main__':
    unittest.main()




"""
DataOutput - class to handle disk file storage
"""
# package to handle files/folders and related metadata/operations
import os
# package facilitating Data Frames manipulation
import pandas


class DataDiskRead:

    @staticmethod
    def fn_internal_load_csv_file_into_data_frame(in_dict):
        if in_dict['format'].lower() == 'csv':
            try:
                out_data_frame = []
                for index_file, crt_file in enumerate(in_dict['files list']):
                    out_data_frame.append(index_file)
                    out_data_frame[index_file] = pandas.read_csv(
                        filepath_or_buffer=crt_file, delimiter=in_dict['field delimiter'],
                        cache_dates=True, index_col=None, memory_map=True, low_memory=False,
                        encoding='utf-8')
                    out_data_frame[index_file]['Source Data File Name'] = os.path.basename(crt_file)
                in_dict['out data frame'] = pandas.concat(out_data_frame)
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_load_excel_file_into_data_frame(in_dict):
        if in_dict['format'].lower() == 'excel':
            try:
                out_data_frame = []
                for index_file, crt_file in enumerate(in_dict['files list']):
                    out_data_frame.append(index_file)
                    out_data_frame[index_file] = pandas.read_excel(
                        io=crt_file, sheet_name=in_dict['worksheet list'][0], verbose=False)
                    out_data_frame[index_file]['Source Data File Name'] = os.path.basename(crt_file)
                in_dict['out data frame'] = pandas.concat(out_data_frame)
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_load_json_file_into_data_frame(in_dict):
        if in_dict['format'].lower() == 'json':
            try:
                out_data_frame = []
                for index_file, crt_file in enumerate(in_dict['files list']):
                    out_data_frame.append(index_file)
                    out_data_frame[index_file] = pandas.read_json(
                        path_or_buf=crt_file, compression=in_dict['compression'])
                    out_data_frame[index_file]['Source Data File Name'] = os.path.basename(crt_file)
                in_dict['out data frame'] = pandas.concat(out_data_frame)
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_load_parquet_file_into_data_frame(in_dict):
        if in_dict['format'].lower() == 'parquet':
            try:
                out_data_frame = []
                for index_file, crt_file in enumerate(in_dict['files list']):
                    out_data_frame.append(index_file)
                    out_data_frame[index_file] = pandas.read_parquet(path=crt_file)
                    out_data_frame[index_file]['Source Data File Name'] = os.path.basename(crt_file)
                in_dict['out data frame'] = pandas.concat(out_data_frame)
            except Exception as err:
                in_dict['error details'] = err
        return in_dict

    @staticmethod
    def fn_internal_load_pickle_file_into_data_frame(in_dict):
        if in_dict['format'].lower() == 'pickle':
            try:
                out_data_frame = []
                for index_file, crt_file in enumerate(in_dict['files list']):
                    out_data_frame.append(index_file)
                    out_data_frame[index_file] = pandas.read_pickle(
                        filepath_or_buffer=crt_file, compression=in_dict['compression'])
                    out_data_frame[index_file]['Source Data File Name'] = os.path.basename(crt_file)
                in_dict['out data frame'] = pandas.concat(out_data_frame)
            except Exception as err:
                in_dict['error details'] = err
        return in_dict



"""
main - entry point of the package

This file is performing CSV read into HYPER file and measures time elapsed (performance)
"""
# package to handle files/folders and related metadata/operations
import os
# Custom classes specific to this package
from project_locale.localizations_common import LocalizationsCommon
from tableau_hyper_management.ProjectNeeds import ProjectNeeds
from tableau_hyper_management.TableauHyperApiExtraLogic import TableauHyperApiExtraLogic
from tableau_hyper_management.TypeDetermination import TypeDetermination
# get current script name
SCRIPT_NAME = os.path.basename(__file__).replace('.py', '')

# main execution logic
if __name__ == '__main__':
    # instantiate Localizations Common class
    class_lc = LocalizationsCommon()
    # ensure all localization templates files are older than translated files
    class_lc.run_localization_action('maintain_sources')
    # ensure all compiled localization files are in place (as needed for localized messages later)
    class_lc.run_localization_action('compile')
    # establish localization language to use
    language_to_use = class_lc.get_region_language_to_use_from_operating_system()
    # instantiate Extractor Specific Needs class
    class_pn = ProjectNeeds({
        'script name': SCRIPT_NAME,
        'language': language_to_use,
        'title': 'Tableau Hyper Converter'
    })
    # as input and/or output file might contain CalculatedDate expression
    # an evaluation is required
    class_pn.parameters.input_file = class_pn.class_ph.eval_expression(
        class_pn.class_ln.logger, class_pn.parameters.input_file, 7)
    class_pn.parameters.output_file = class_pn.class_ph.eval_expression(
        class_pn.class_ln.logger, class_pn.parameters.output_file, 7)
    # as destination folder could be dynamic and not existent safety measure is required
    destination_folder = os.path.dirname(class_pn.parameters.output_file)
    if not os.path.exists(destination_folder):
        os.makedirs(destination_folder)
    # check if proceeding is necessary
    load_data_frame_necessary = True
    if class_pn.parameters.input_file_format != 'hyper' \
        and class_pn.parameters.output_file_format == 'hyper' \
            and class_pn.parameters.policy_to_handle_hyper_file == 'create':
        if os.path.isfile(class_pn.parameters.output_file):
            # initial assumption is no further load is necessary
            load_data_frame_necessary = False
            relevant_files_list = []
            feedback = [
                'The output file defined as {file_name} already exists',
                'As handle hyper action is set to "create", '
                + 'further checks will happen to ensure none of the source files '
                + 'is newer than destination',
            ]
            for crt_feedback in feedback:
                class_pn.class_ln.logger.error(class_pn.locale.gettext(
                    crt_feedback.replace('{file_name}', class_pn.parameters.output_file)))
                class_pn.class_bn.fn_timestamped_print(class_pn.locale.gettext(
                    crt_feedback.replace('{file_name}', class_pn.parameters.output_file)))
    # Even if the Hyper file exists already, can happen one of the source file is newer
    # so an overwrite might be more appropriate in such case
    # identify all files matching input file information
    relevant_files_list = class_pn.class_fo.fn_build_file_list(
        class_pn.class_ln.logger, class_pn.timer, class_pn.parameters.input_file)
    # log file statistic details
    class_pn.class_fo.fn_store_file_statistics({
            'file list': relevant_files_list,
            'file meaning': 'Input',
            'checksum included': class_pn.parameters.include_checksum_in_files_statistics,
            'logger': class_pn.class_ln.logger,
            'timer': class_pn.timer,
    })
    # further could be required to assess "load_data_frame_necessary" value
    if not load_data_frame_necessary:
        final_verdict = class_pn.source_vs_destination_file_modification_assesment(
            class_pn.class_ln.logger, class_pn.timer, {
                'destination file': class_pn.parameters.output_file,
                'list source files': relevant_files_list,
            })
        # as final_verdict values could be different depending on localization used/detected
        # only a check of None or not None is possible
        if final_verdict is not None:
            load_data_frame_necessary = True
            class_pn.parameters.policy_to_handle_hyper_file = 'overwrite'
    # instantiate Tableau Hyper Api Extra Logic class
    class_thael = TableauHyperApiExtraLogic(language_to_use)
    # loading from a specific folder all files matching a given pattern into a data frame
    input_dict = {
        'action': class_pn.parameters.policy_to_handle_hyper_file,
        'compression': class_pn.parameters.input_file_compression,
        'field delimiter': class_pn.parameters.csv_field_separator,
        'file list': relevant_files_list,
        'format': class_pn.parameters.input_file_format,
        'name': 'irrelevant',
        'query': class_pn.parameters.sql_query_to_handle_data,
        'schema name': 'Extract',
        'table name': 'Extract',
    }
    working_data_frame = None
    if class_pn.parameters.input_file_format == 'hyper':
        if relevant_files_list:
            input_dict['action'] = 'read from existing hyper'
            input_dict['hyper file'] = relevant_files_list[0]
            working_data_frame = class_thael.fn_hyper_handle(
                class_pn.class_ln.logger, class_pn.timer, input_dict)
    elif load_data_frame_necessary:
        working_data_frame = class_pn.class_dio.fn_load_file_into_data_frame(
            class_pn.class_ln.logger, class_pn.timer, input_dict)
    if working_data_frame is not None:
        output_dict = input_dict
        # overwrite few important values with relevant information for output
        output_dict['file list'] = 'irrelevant'
        output_dict['format'] = class_pn.parameters.output_file_format
        output_dict['name'] = class_pn.parameters.output_file
        output_dict['compression'] = class_pn.parameters.output_file_compression
        if class_pn.parameters.input_file_format.lower() == 'hyper':
            tuple_supported_file_types = class_thael.supported_output_file_types
        else:
            tuple_supported_file_types = class_pn.class_dio.implemented_disk_write_file_types
        wanted_output_format = class_pn.parameters.output_file_format.lower()
        if class_pn.parameters.output_file_format.lower() in tuple_supported_file_types:
            class_pn.class_dio.fn_store_data_frame_to_file(
                class_pn.class_ln.logger, class_pn.timer, working_data_frame, output_dict)
            # store statistics about output file
            class_pn.class_fo.fn_store_file_statistics({
                'file list': class_pn.parameters.output_file,
                'file meaning': 'Generated',
                'checksum included': class_pn.parameters.include_checksum_in_files_statistics,
                'logger': class_pn.class_ln.logger,
                'timer': class_pn.timer,
            })
        elif wanted_output_format == 'hyper':
            supported_types = class_thael.supported_input_file_types
            if class_pn.parameters.input_file_format.lower() in supported_types:
                c_td = TypeDetermination(language_to_use)
                fn_dict = {
                    'action': input_dict['action'],
                    'data frame': working_data_frame,
                    'input parameters': class_pn.parameters,
                    'input data types': class_pn.config['data_types'],
                    'hyper file': class_pn.parameters.output_file,
                    'schema name': input_dict['schema name'],
                    'table name': input_dict['table name'],
                }
                if fn_dict['action'] in ('append', 'create', 'overwrite'):
                    # advanced detection of data type within Data Frame
                    fn_dict['data frame structure'] = c_td.fn_get_data_frame_structure(
                        class_pn.class_ln.logger, class_pn.timer, fn_dict)
                    # determine Hyper Table Columns
                    fn_dict['hyper table columns'] = class_thael.fn_build_hyper_columns(
                        class_pn.class_ln.logger, class_pn.timer, fn_dict['data frame structure'])
                    # The rows to insert into the <hyper_table> table.
                    fn_dict['data'] = class_thael.fn_rebuild_data_frame_content_for_hyper(
                        class_pn.class_ln.logger, class_pn.timer, fn_dict)
                    # check if output Hyper file does not exists
                    # and if so action will be always "overwrite"
                    # which will trigger internal Hyper structure creation (schema and table)
                    if not os.path.isfile(fn_dict['hyper file']):
                        fn_dict['action'] = 'overwrite'
                # manipulate destination Tableau Extract (Hyper)
                class_thael.fn_hyper_handle(class_pn.class_ln.logger, class_pn.timer, fn_dict)
                # store statistics about output file
                class_pn.class_fo.fn_store_file_statistics({
                    'file list': class_pn.parameters.output_file,
                    'file meaning': 'Generated',
                    'checksum included': class_pn.parameters.include_checksum_in_files_statistics,
                    'logger': class_pn.class_ln.logger,
                    'timer': class_pn.timer,
                })
            else:
                class_pn.class_ln.logger.error(
                    class_pn.locale.gettext(
                        'Only certain file types are supported as input file type '
                        + 'in combination with Tableau Extract (Hyper) as output file type. '
                        + 'And {given_file_type} is not among "{supported_file_types}"')
                        .replace('{given_file_type}', wanted_output_format)
                        .replace('{supported_file_types}', '", "'.join(supported_types)))
    # just final message
    class_pn.class_bn.fn_final_message(
        class_pn.class_ln.logger, class_pn.parameters.output_log_file,
        class_pn.timer.timers.total(SCRIPT_NAME))
